import asyncio
import logging
from typing import Any, TypedDict

import socketio  # type: ignore[import-untyped]

from pydase.data_service.data_service import process_callable_attribute
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.data_service.state_manager import StateManager
from pydase.utils.helpers import get_object_attr_from_path_list
from pydase.utils.logging import SocketIOHandler
from pydase.utils.serializer import dump

logger = logging.getLogger(__name__)


class UpdateDict(TypedDict):
    """
    A TypedDict subclass representing a dictionary used for updating attributes in a
    DataService.

    Attributes:
    ----------
    name : str
        The name of the attribute to be updated in the DataService instance.
        If the attribute is part of a nested structure, this would be the name of the
        attribute in the last nested object. For example, for an attribute access path
        'attr1.list_attr[0].attr2', 'attr2' would be the name.

    parent_path : str
        The access path for the parent object of the attribute to be updated. This is
        used to construct the full access path for the attribute. For example, for an
        attribute access path 'attr1.list_attr[0].attr2', 'attr1.list_attr[0]' would be
        the parent_path.

    value : Any
        The new value to be assigned to the attribute. The type of this value should
        match the type of the attribute to be updated.
    """

    name: str
    parent_path: str
    value: Any


class RunMethodDict(TypedDict):
    """
    A TypedDict subclass representing a dictionary used for running methods from the
    exposed DataService.

    Attributes:
        name (str): The name of the method to be run.
        parent_path (str): The access path for the parent object of the method to be
            run. This is used to construct the full access path for the method. For
            example, for an method with access path 'attr1.list_attr[0].method_name',
            'attr1.list_attr[0]' would be the parent_path.
        kwargs (dict[str, Any]): The arguments passed to the method.
    """

    name: str
    parent_path: str
    kwargs: dict[str, Any]


def setup_sio_server(
    observer: DataServiceObserver,
    enable_cors: bool,
    loop: asyncio.AbstractEventLoop,
) -> socketio.AsyncServer:
    """
    Sets up and configures a Socket.IO asynchronous server.

    Args:
        observer (DataServiceObserver):
          The observer managing state updates and communication.
        enable_cors (bool):
          Flag indicating whether CORS should be enabled for the server.
        loop (asyncio.AbstractEventLoop):
          The event loop in which the server will run.

    Returns:
        socketio.AsyncServer: The configured Socket.IO asynchronous server.
    """

    state_manager = observer.state_manager

    if enable_cors:
        sio = socketio.AsyncServer(async_mode="asgi", cors_allowed_origins="*")
    else:
        sio = socketio.AsyncServer(async_mode="asgi")

    setup_sio_events(sio, state_manager)
    setup_logging_handler(sio)

    # Add notification callback to observer
    def sio_callback(
        full_access_path: str, value: Any, cached_value_dict: dict[str, Any]
    ) -> None:
        if cached_value_dict != {}:
            serialized_value = dump(value)
            if cached_value_dict["type"] != "method":
                cached_value_dict["type"] = serialized_value["type"]

            cached_value_dict["value"] = serialized_value["value"]

            async def notify() -> None:
                try:
                    await sio.emit(
                        "notify",
                        {
                            "data": {
                                "full_access_path": full_access_path,
                                "value": cached_value_dict,
                            }
                        },
                    )
                except Exception as e:
                    logger.warning("Failed to send notification: %s", e)

            loop.create_task(notify())

    observer.add_notification_callback(sio_callback)

    return sio


def setup_sio_events(sio: socketio.AsyncServer, state_manager: StateManager) -> None:
    @sio.event
    def set_attribute(sid: str, data: UpdateDict) -> Any:
        logger.debug("Received frontend update: %s", data)
        parent_path = data["parent_path"].split(".")
        path_list = [element for element in parent_path if element] + [data["name"]]
        path = ".".join(path_list)
        return state_manager.set_service_attribute_value_by_path(
            path=path, value=data["value"]
        )

    @sio.event
    def run_method(sid: str, data: RunMethodDict) -> Any:
        logger.debug("Running method: %s", data)
        parent_path = data["parent_path"].split(".")
        path_list = [element for element in parent_path if element] + [data["name"]]
        method = get_object_attr_from_path_list(state_manager.service, path_list)
        return process_callable_attribute(method, data["kwargs"])


def setup_logging_handler(sio: socketio.AsyncServer) -> None:
    logger = logging.getLogger()
    logger.addHandler(SocketIOHandler(sio))



import asyncio
import logging
import os
import signal
import threading
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from types import FrameType
from typing import Any, Protocol, TypedDict

from rpyc import ThreadedServer  # type: ignore[import-untyped]
from uvicorn.server import HANDLED_SIGNALS

from pydase import DataService
from pydase.config import ServiceConfig
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.data_service.state_manager import StateManager
from pydase.server.web_server import WebServer

logger = logging.getLogger(__name__)


class AdditionalServerProtocol(Protocol):
    """
    A Protocol that defines the interface for additional servers.

    This protocol sets the standard for how additional servers should be implemented
    to ensure compatibility with the main Server class. The protocol requires that
    any server implementing it should have an __init__ method for initialization and a
    serve method for starting the server.

    Args:
        data_service_observer:
          Observer for the DataService, handling state updates and communication to
          connected clients through injected callbacks. Can be utilized to access the
          service and state manager, and to add custom state-update callbacks.
        host:
          Hostname or IP address where the server is accessible. Commonly '0.0.0.0' to
          bind to all network interfaces.
        port:
          Port number on which the server listens. Typically in the range 1024-65535
          (non-standard ports).
        **kwargs:
          Any additional parameters required for initializing the server. These
          parameters are specific to the server's implementation.
    """

    def __init__(
        self,
        data_service_observer: DataServiceObserver,
        host: str,
        port: int,
        **kwargs: Any,
    ) -> None:
        ...

    async def serve(self) -> Any:
        """Starts the server. This method should be implemented as an asynchronous
        method, which means that it should be able to run concurrently with other tasks.
        """


class AdditionalServer(TypedDict):
    """
    A TypedDict that represents the configuration for an additional server to be run
    alongside the main server.

    This class is used to specify the server type, the port on which the server should
    run, and any additional keyword arguments that should be passed to the server when
    it's instantiated.
    """

    server: type[AdditionalServerProtocol]
    port: int
    kwargs: dict[str, Any]


class Server:
    """
    The `Server` class provides a flexible server implementation for the `DataService`.

    Args:
        service: DataService
          The DataService instance that this server will manage.
        host: str
          The host address for the server. Default is '0.0.0.0', which means all
          available network interfaces.
        rpc_port: int
          The port number for the RPC server. Default is
          `pydase.config.ServiceConfig().rpc_port`.
        web_port: int
          The port number for the web server. Default is
          `pydase.config.ServiceConfig().web_port`.
        enable_rpc: bool
          Whether to enable the RPC server. Default is True.
        enable_web: bool
          Whether to enable the web server. Default is True.
        filename: str | Path | None
          Filename of the file managing the service state persistence. Defaults to None.
        use_forking_server: bool
          Whether to use ForkingServer for multiprocessing. Default is False.
        additional_servers : list[AdditionalServer]
          A list of additional servers to run alongside the main server. Each entry in
          the list should be a dictionary with the following structure:
            - server: A class that adheres to the AdditionalServerProtocol. This class
                should have an `__init__` method that accepts the DataService instance,
                port, host, and optional keyword arguments, and a `serve` method that is
                a coroutine responsible for starting the server.
            - port: The port on which the additional server will be running.
            - kwargs: A dictionary containing additional keyword arguments that will be
                passed to the server's `__init__` method.

          Here's an example of how you might define an additional server:


          >>>     class MyCustomServer:
          ...         def __init__(
          ...             self,
          ...             data_service_observer: DataServiceObserver,
          ...             host: str,
          ...             port: int,
          ...             **kwargs: Any,
          ...         ) -> None:
          ...             self.observer = data_service_observer
          ...             self.state_manager = self.observer.state_manager
          ...             self.service = self.state_manager.service
          ...             self.port = port
          ...             self.host = host
          ...             # handle any additional arguments...
          ...
          ...         async def serve(self):
          ...             # code to start the server...

          And here's how you might add it to the `additional_servers` list when creating
          a `Server` instance:

          >>>    server = Server(
          ...        service=my_data_service,
          ...        additional_servers=[
          ...            {
          ...                "server": MyCustomServer,
          ...                "port": 12345,
          ...                "kwargs": {"some_arg": "some_value"}
          ...            }
          ...        ],
          ...    )
          ...    server.run()

        **kwargs: Any
          Additional keyword arguments.
    """

    def __init__(  # noqa: PLR0913
        self,
        service: DataService,
        host: str = "0.0.0.0",
        rpc_port: int = ServiceConfig().rpc_port,
        web_port: int = ServiceConfig().web_port,
        enable_rpc: bool = True,
        enable_web: bool = True,
        filename: str | Path | None = None,
        additional_servers: list[AdditionalServer] | None = None,
        **kwargs: Any,
    ) -> None:
        if additional_servers is None:
            additional_servers = []
        self._service = service
        self._host = host
        self._rpc_port = rpc_port
        self._web_port = web_port
        self._enable_rpc = enable_rpc
        self._enable_web = enable_web
        self._kwargs = kwargs
        self._loop: asyncio.AbstractEventLoop
        self._additional_servers = additional_servers
        self.should_exit = False
        self.servers: dict[str, asyncio.Future[Any]] = {}
        self.executor: ThreadPoolExecutor | None = None
        self._state_manager = StateManager(self._service, filename)
        self._observer = DataServiceObserver(self._state_manager)
        self._state_manager.load_state()

    def run(self) -> None:
        """
        Initializes the asyncio event loop and starts the server.

        This method should be called to start the server after it's been instantiated.
        """
        asyncio.run(self.serve())

    async def serve(self) -> None:
        process_id = os.getpid()

        logger.info("Started server process [%s]", process_id)

        await self.startup()
        if self.should_exit:
            return
        await self.main_loop()
        await self.shutdown()

        logger.info("Finished server process [%s]", process_id)

    async def startup(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._loop.set_exception_handler(self.custom_exception_handler)
        self.install_signal_handlers()
        self._service._task_manager.start_autostart_tasks()

        if self._enable_rpc:
            self.executor = ThreadPoolExecutor()
            self._rpc_server = ThreadedServer(
                self._service,
                port=self._rpc_port,
                protocol_config={
                    "allow_all_attrs": True,
                    "allow_setattr": True,
                },
            )
            future_or_task = self._loop.run_in_executor(
                executor=self.executor, func=self._rpc_server.start
            )
            self.servers["rpyc"] = future_or_task
        for server in self._additional_servers:
            addin_server = server["server"](
                data_service_observer=self._observer,
                host=self._host,
                port=server["port"],
                **server["kwargs"],
            )

            server_name = (
                addin_server.__module__ + "." + addin_server.__class__.__name__
            )

            future_or_task = self._loop.create_task(addin_server.serve())
            self.servers[server_name] = future_or_task
        if self._enable_web:
            self._web_server = WebServer(
                data_service_observer=self._observer,
                host=self._host,
                port=self._web_port,
                **self._kwargs,
            )
            future_or_task = self._loop.create_task(self._web_server.serve())
            self.servers["web"] = future_or_task

    async def main_loop(self) -> None:
        while not self.should_exit:
            await asyncio.sleep(0.1)

    async def shutdown(self) -> None:
        logger.info("Shutting down")

        logger.info("Saving data to %s.", self._state_manager.filename)
        self._state_manager.save_state()

        await self.__cancel_servers()
        await self.__cancel_tasks()

        if hasattr(self, "_rpc_server") and self._enable_rpc:
            logger.debug("Closing rpyc server.")
            self._rpc_server.close()

    async def __cancel_servers(self) -> None:
        for server_name, task in self.servers.items():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.debug("Cancelled '%s' server.", server_name)
            except Exception as e:
                logger.warning("Unexpected exception: %s", e)

    async def __cancel_tasks(self) -> None:
        for task in asyncio.all_tasks(self._loop):
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.debug("Cancelled task '%s'.", task.get_coro())
            except Exception as e:
                logger.exception("Unexpected exception: %s", e)

    def install_signal_handlers(self) -> None:
        if threading.current_thread() is not threading.main_thread():
            # Signals can only be listened to from the main thread.
            return

        for sig in HANDLED_SIGNALS:
            signal.signal(sig, self.handle_exit)

    def handle_exit(self, sig: int = 0, frame: FrameType | None = None) -> None:
        if self.should_exit and sig == signal.SIGINT:
            logger.warning("Received signal '%s', forcing exit...", sig)
            os._exit(1)
        else:
            self.should_exit = True
            logger.warning(
                "Received signal '%s', exiting... (CTRL+C to force quit)", sig
            )

    def custom_exception_handler(
        self, loop: asyncio.AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        # if any background task creates an unhandled exception, shut down the entire
        # loop. It's possible we don't want to do this, maybe make this optional in the
        # future
        loop.default_exception_handler(context)

        # here we exclude most kinds of exceptions from triggering this kind of shutdown
        exc = context.get("exception")
        if type(exc) not in [RuntimeError, KeyboardInterrupt, asyncio.CancelledError]:
            if self._enable_web:

                async def emit_exception() -> None:
                    try:
                        await self._web_server._sio.emit(
                            "exception",
                            {
                                "data": {
                                    "exception": str(exc),
                                    "type": exc.__class__.__name__,
                                }
                            },
                        )
                    except Exception as e:
                        logger.exception("Failed to send notification: %s", e)

                loop.create_task(emit_exception())
        else:
            self.handle_exit()



import inspect
import logging
import sys
from collections.abc import Callable
from enum import Enum
from typing import Any, TypedDict

import pydase.units as u
from pydase.data_service.abstract_data_service import AbstractDataService
from pydase.data_service.task_manager import TaskStatus
from pydase.utils.helpers import (
    get_attribute_doc,
    get_component_classes,
    get_data_service_class_reference,
    parse_list_attr_and_index,
    render_in_frontend,
)

logger = logging.getLogger(__name__)


class SerializationPathError(Exception):
    pass


class SerializationValueError(Exception):
    pass


class Serializer:
    @staticmethod
    def serialize_object(obj: Any) -> dict[str, Any]:
        result: dict[str, Any] = {}
        if isinstance(obj, AbstractDataService):
            result = Serializer._serialize_data_service(obj)

        elif isinstance(obj, list):
            result = Serializer._serialize_list(obj)

        elif isinstance(obj, dict):
            result = Serializer._serialize_dict(obj)

        # Special handling for u.Quantity
        elif isinstance(obj, u.Quantity):
            result = Serializer._serialize_quantity(obj)

        # Handling for Enums
        elif isinstance(obj, Enum):
            result = Serializer._serialize_enum(obj)

        # Methods and coroutines
        elif inspect.isfunction(obj) or inspect.ismethod(obj):
            result = Serializer._serialize_method(obj)

        else:
            obj_type = type(obj).__name__
            value = obj
            readonly = False
            doc = get_attribute_doc(obj)
            result = {
                "type": obj_type,
                "value": value,
                "readonly": readonly,
                "doc": doc,
            }

        return result

    @staticmethod
    def _serialize_enum(obj: Enum) -> dict[str, Any]:
        import pydase.components.coloured_enum

        value = obj.name
        readonly = False
        doc = obj.__doc__
        if sys.version_info < (3, 11) and doc == "An enumeration.":
            doc = None
        if isinstance(obj, pydase.components.coloured_enum.ColouredEnum):
            obj_type = "ColouredEnum"
        else:
            obj_type = "Enum"

        return {
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
            "enum": {
                name: member.value for name, member in obj.__class__.__members__.items()
            },
        }

    @staticmethod
    def _serialize_quantity(obj: u.Quantity) -> dict[str, Any]:
        obj_type = "Quantity"
        readonly = False
        doc = get_attribute_doc(obj)
        value = {"magnitude": obj.m, "unit": str(obj.u)}
        return {
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    @staticmethod
    def _serialize_dict(obj: dict[str, Any]) -> dict[str, Any]:
        obj_type = "dict"
        readonly = False
        doc = get_attribute_doc(obj)
        value = {key: Serializer.serialize_object(val) for key, val in obj.items()}
        return {
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    @staticmethod
    def _serialize_list(obj: list[Any]) -> dict[str, Any]:
        obj_type = "list"
        readonly = False
        doc = get_attribute_doc(obj)
        value = [Serializer.serialize_object(o) for o in obj]
        return {
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    @staticmethod
    def _serialize_method(obj: Callable[..., Any]) -> dict[str, Any]:
        obj_type = "method"
        value = None
        readonly = True
        doc = get_attribute_doc(obj)
        frontend_render = render_in_frontend(obj)

        # Store parameters and their anotations in a dictionary
        sig = inspect.signature(obj)
        sig.return_annotation

        class SignatureDict(TypedDict):
            parameters: dict[str, dict[str, Any]]
            return_annotation: dict[str, Any]

        signature: SignatureDict = {"parameters": {}, "return_annotation": {}}

        for k, v in sig.parameters.items():
            signature["parameters"][k] = {
                "annotation": str(v.annotation),
                "default": dump(v.default) if v.default != inspect._empty else {},
            }

        return {
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
            "async": inspect.iscoroutinefunction(obj),
            "signature": signature,
            "frontend_render": frontend_render,
        }

    @staticmethod
    def _serialize_data_service(obj: AbstractDataService) -> dict[str, Any]:
        readonly = False
        doc = get_attribute_doc(obj)
        obj_type = "DataService"
        obj_name = obj.__class__.__name__

        # Get component base class if any
        component_base_cls = next(
            (cls for cls in get_component_classes() if isinstance(obj, cls)), None
        )
        if component_base_cls:
            obj_type = component_base_cls.__name__

        # Get the set of DataService class attributes
        data_service_attr_set = set(dir(get_data_service_class_reference()))
        # Get the set of the object attributes
        obj_attr_set = set(dir(obj))
        # Get the difference between the two sets
        derived_only_attr_set = obj_attr_set - data_service_attr_set

        value = {}

        # Iterate over attributes, properties, class attributes, and methods
        for key in sorted(derived_only_attr_set):
            if key.startswith("_"):
                continue  # Skip attributes that start with underscore

            # Skip keys that start with "start_" or "stop_" and end with an async
            # method name
            if key.startswith(("start_", "stop_")) and key.split("_", 1)[1] in {
                name
                for name, _ in inspect.getmembers(
                    obj, predicate=inspect.iscoroutinefunction
                )
            }:
                continue

            val = getattr(obj, key)

            value[key] = Serializer.serialize_object(val)

            # If there's a running task for this method
            if key in obj._task_manager.tasks:
                value[key]["value"] = TaskStatus.RUNNING.name

            # If the DataService attribute is a property
            if isinstance(getattr(obj.__class__, key, None), property):
                prop: property = getattr(obj.__class__, key)
                value[key]["readonly"] = prop.fset is None
                value[key]["doc"] = get_attribute_doc(prop)  # overwrite the doc

        return {
            "name": obj_name,
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }


def dump(obj: Any) -> dict[str, Any]:
    return Serializer.serialize_object(obj)


def set_nested_value_by_path(
    serialization_dict: dict[str, Any], path: str, value: Any
) -> None:
    """
    Set a value in a nested dictionary structure, which conforms to the serialization
    format used by `pydase.utils.serializer.Serializer`, using a dot-notation path.

    Args:
        serialization_dict:
            The base dictionary representing data serialized with
            `pydase.utils.serializer.Serializer`.
        path:
            The dot-notation path (e.g., 'attr1.attr2[0].attr3') indicating where to
            set the value.
        value:
            The new value to set at the specified path.

    Note:
        - If the index equals the length of the list, the function will append the
          serialized representation of the 'value' to the list.
    """

    parent_path_parts, attr_name = path.split(".")[:-1], path.split(".")[-1]
    current_dict: dict[str, Any] = serialization_dict

    try:
        for path_part in parent_path_parts:
            current_dict = get_next_level_dict_by_key(
                current_dict, path_part, allow_append=False
            )
            current_dict = current_dict["value"]

        current_dict = get_next_level_dict_by_key(
            current_dict, attr_name, allow_append=True
        )
    except (SerializationPathError, SerializationValueError, KeyError) as e:
        logger.error(e)
        return

    # setting the new value
    serialized_value = dump(value)
    if "readonly" in current_dict:
        if current_dict["type"] != "method":
            current_dict["type"] = serialized_value["type"]
        current_dict["value"] = serialized_value["value"]
    else:
        current_dict.update(serialized_value)


def get_nested_dict_by_path(
    serialization_dict: dict[str, Any],
    path: str,
) -> dict[str, Any]:
    parent_path_parts, attr_name = path.split(".")[:-1], path.split(".")[-1]
    current_dict: dict[str, Any] = serialization_dict

    for path_part in parent_path_parts:
        current_dict = get_next_level_dict_by_key(
            current_dict, path_part, allow_append=False
        )
        current_dict = current_dict["value"]
    return get_next_level_dict_by_key(current_dict, attr_name, allow_append=False)


def get_next_level_dict_by_key(
    serialization_dict: dict[str, Any],
    attr_name: str,
    *,
    allow_append: bool = False,
) -> dict[str, Any]:
    """
    Retrieve a nested dictionary entry or list item from a data structure serialized
    with `pydase.utils.serializer.Serializer`.

    Args:
        serialization_dict: The base dictionary representing serialized data.
        attr_name: The key name representing the attribute in the dictionary,
            e.g. 'list_attr[0]' or 'attr'
        allow_append: Flag to allow appending a new entry if `index` is out of range by
            one.

    Returns:
        The dictionary or list item corresponding to the attribute and index.

    Raises:
        SerializationPathError: If the path composed of `attr_name` and `index` is
                                invalid or leads to an IndexError or KeyError.
        SerializationValueError: If the expected nested structure is not a dictionary.
    """
    # Check if the key contains an index part like 'attr_name[<index>]'
    attr_name, index = parse_list_attr_and_index(attr_name)

    try:
        if index is not None:
            serialization_dict = serialization_dict[attr_name]["value"][index]
        else:
            serialization_dict = serialization_dict[attr_name]
    except IndexError as e:
        if allow_append and index == len(serialization_dict[attr_name]["value"]):
            # Appending to list
            serialization_dict[attr_name]["value"].append({})
            serialization_dict = serialization_dict[attr_name]["value"][index]
        else:
            raise SerializationPathError(
                f"Error occured trying to change '{attr_name}[{index}]': {e}"
            )
    except KeyError:
        raise SerializationPathError(
            f"Error occured trying to access the key '{attr_name}': it is either "
            "not present in the current dictionary or its value does not contain "
            "a 'value' key."
        )

    if not isinstance(serialization_dict, dict):
        raise SerializationValueError(
            f"Expected a dictionary at '{attr_name}', but found type "
            f"'{type(serialization_dict).__name__}' instead."
        )

    return serialization_dict


def generate_serialized_data_paths(
    data: dict[str, dict[str, Any]], parent_path: str = ""
) -> list[str]:
    """
    Generate a list of access paths for all attributes in a dictionary representing
    data serialized with `pydase.utils.serializer.Serializer`, excluding those that are
    methods. This function handles nested structures, including lists, by generating
    paths for each element in the nested lists.

    Args:
        data (dict[str, Any]): The dictionary representing serialized data, typically
            produced by `pydase.utils.serializer.Serializer`.
        parent_path (str, optional): The base path to prepend to the keys in the `data`
            dictionary to form the access paths. Defaults to an empty string.

    Returns:
        list[str]: A list of strings where each string is a dot-notation access path
        to an attribute in the serialized data. For list elements, the path includes
        the index in square brackets.
    """

    paths: list[str] = []
    for key, value in data.items():
        new_path = f"{parent_path}.{key}" if parent_path else key
        paths.append(new_path)
        if serialized_dict_is_nested_object(value):
            if isinstance(value["value"], list):
                for index, item in enumerate(value["value"]):
                    indexed_key_path = f"{new_path}[{index}]"
                    paths.append(indexed_key_path)
                    if serialized_dict_is_nested_object(item):
                        paths.extend(
                            generate_serialized_data_paths(
                                item["value"], indexed_key_path
                            )
                        )
                continue
            paths.extend(generate_serialized_data_paths(value["value"], new_path))
    return paths


def serialized_dict_is_nested_object(serialized_dict: dict[str, Any]) -> bool:
    return (
        serialized_dict["type"] != "Quantity"
        and isinstance(serialized_dict["value"], dict)
    ) or isinstance(serialized_dict["value"], list)



import inspect
import logging
from collections.abc import Callable
from itertools import chain
from typing import Any

logger = logging.getLogger(__name__)


def get_attribute_doc(attr: Any) -> str | None:
    """This function takes an input attribute attr and returns its documentation
    string if it's different from the documentation of its type, otherwise,
    it returns None.
    """
    attr_doc = inspect.getdoc(attr)
    attr_class_doc = inspect.getdoc(type(attr))
    return attr_doc if attr_class_doc != attr_doc else None


def get_class_and_instance_attributes(obj: object) -> dict[str, Any]:
    """Dictionary containing all attributes (both instance and class level) of a
    given object.

    If an attribute exists at both the instance and class level,the value from the
    instance attribute takes precedence.
    The __root__ object is removed as this will lead to endless recursion in the for
    loops.
    """

    return dict(chain(type(obj).__dict__.items(), obj.__dict__.items()))


def get_object_attr_from_path_list(target_obj: Any, path: list[str]) -> Any:
    """
    Traverse the object tree according to the given path.

    Args:
        target_obj: The root object to start the traversal from.
        path: A list of attribute names representing the path to traverse.

    Returns:
        The attribute at the end of the path. If the path includes a list index,
        the function returns the specific item at that index. If an attribute in
        the path does not exist, the function logs a debug message and returns None.

    Raises:
        ValueError: If a list index in the path is not a valid integer.
    """
    for part in path:
        try:
            # Try to split the part into attribute and index
            attr, index_str = part.split("[", maxsplit=1)
            index_str = index_str.replace("]", "")
            index = int(index_str)
            target_obj = getattr(target_obj, attr)[index]
        except ValueError:
            # No index, so just get the attribute
            target_obj = getattr(target_obj, part)
        except AttributeError:
            # The attribute doesn't exist
            logger.debug("Attribute % does not exist in the object.", part)
            return None
    return target_obj


def convert_arguments_to_hinted_types(
    args: dict[str, Any], type_hints: dict[str, Any]
) -> dict[str, Any] | str:
    """
    Convert the given arguments to their types hinted in the type_hints dictionary.

    This function attempts to convert each argument in the args dictionary to the type
    specified for the argument in the type_hints dictionary. If the conversion is
    successful, the function replaces the original argument in the args dictionary with
    the converted argument.

    If a ValueError is raised during the conversion of an argument, the function logs
    an error message and returns the error message as a string.

    Args:
        args: A dictionary of arguments to be converted. The keys are argument names
              and the values are the arguments themselves.
        type_hints: A dictionary of type hints for the arguments. The keys are
                    argument names and the values are the hinted types.

    Returns:
        A dictionary of the converted arguments if all conversions are successful,
        or an error message string if a ValueError is raised during a conversion.
    """

    # Convert arguments to their hinted types
    for arg_name, arg_value in args.items():
        if arg_name in type_hints:
            arg_type = type_hints[arg_name]
            if isinstance(arg_type, type):
                # Attempt to convert the argument to its hinted type
                try:
                    args[arg_name] = arg_type(arg_value)
                except ValueError:
                    msg = (
                        f"Failed to convert argument '{arg_name}' to type "
                        f"{arg_type.__name__}"
                    )
                    logger.error(msg)
                    return msg
    return args


def update_value_if_changed(
    target: Any, attr_name_or_index: str | int, new_value: Any
) -> None:
    """
    Updates the value of an attribute or a list element on a target object if the new
    value differs from the current one.

    This function supports updating both attributes of an object and elements of a list.

    - For objects, the function first checks the current value of the attribute. If the
      current value differs from the new value, the function updates the attribute.

    - For lists, the function checks the current value at the specified index. If the
      current value differs from the new value, the function updates the list element
      at the given index.

    Args:
        target (Any):
            The target object that has the attribute or the list.
        attr_name_or_index (str | int):
            The name of the attribute or the index of the list element.
        new_value (Any):
            The new value for the attribute or the list element.
    """

    if isinstance(target, list) and isinstance(attr_name_or_index, int):
        if target[attr_name_or_index] != new_value:
            target[attr_name_or_index] = new_value
    elif isinstance(attr_name_or_index, str):
        # If the type matches and the current value is different from the new value,
        # update the attribute.
        if getattr(target, attr_name_or_index) != new_value:
            setattr(target, attr_name_or_index, new_value)
    else:
        logger.error("Incompatible arguments: %s, %s.", target, attr_name_or_index)


def parse_list_attr_and_index(attr_string: str) -> tuple[str, int | None]:
    """
    Parses an attribute string and extracts a potential list attribute name and its
    index.
    Logs an error if the index is not a valid digit.

    Args:
        attr_string (str):
            The attribute string to parse. Can be a regular attribute name (e.g.,
            'attr_name') or a list attribute with an index (e.g., 'list_attr[2]').

    Returns:
        tuple[str, Optional[int]]:
            A tuple containing the attribute name as a string and the index as an
            integer if present, otherwise None.

    Examples:
        >>> parse_attribute_and_index('list_attr[2]')
        ('list_attr', 2)
        >>> parse_attribute_and_index('attr_name')
        ('attr_name', None)
    """

    index = None
    attr_name = attr_string
    if "[" in attr_string and attr_string.endswith("]"):
        attr_name, index_part = attr_string.split("[", 1)
        index_part = index_part.rstrip("]")
        if index_part.isdigit():
            index = int(index_part)
        else:
            logger.error("Invalid index format in key: %s", attr_name)
    return attr_name, index


def get_component_classes() -> list[type]:
    """
    Returns references to the component classes in a list.
    """
    import pydase.components

    return [
        getattr(pydase.components, cls_name) for cls_name in pydase.components.__all__
    ]


def get_data_service_class_reference() -> Any:
    import pydase.data_service.data_service

    return getattr(pydase.data_service.data_service, "DataService")


def is_property_attribute(target_obj: Any, attr_name: str) -> bool:
    return isinstance(getattr(type(target_obj), attr_name, None), property)


def function_has_arguments(func: Callable[..., Any]) -> bool:
    sig = inspect.signature(func)
    parameters = dict(sig.parameters)
    # Remove 'self' parameter for instance methods.
    parameters.pop("self", None)

    # Check if there are any parameters left which would indicate additional arguments.
    if len(parameters) > 0:
        return True
    return False


def render_in_frontend(func: Callable[..., Any]) -> bool:
    """Determines if the method should be rendered in the frontend.

    It checks if the "@frontend" decorator was used or the method is a coroutine."""

    if inspect.iscoroutinefunction(func):
        return True

    try:
        return func._display_in_frontend  # type: ignore
    except AttributeError:
        return False



from collections.abc import Callable
from typing import Any

from pydase.utils.helpers import function_has_arguments


class FunctionDefinitionError(Exception):
    pass


def frontend(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to mark a DataService method for frontend rendering. Ensures that the
    method does not contain arguments, as they are not supported for frontend rendering.
    """

    if function_has_arguments(func):
        raise FunctionDefinitionError(
            "The @frontend decorator requires functions without arguments. Function "
            f"'{func.__name__}' has at least one argument. "
            "Please remove the argument(s) from this function to use it with the "
            "@frontend decorator."
        )

    # Mark the function for frontend display.
    func._display_in_frontend = True  # type: ignore
    return func



import json
import logging
import os
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import pydase.units as u
from pydase.data_service.data_service_cache import DataServiceCache
from pydase.utils.helpers import (
    get_object_attr_from_path_list,
    is_property_attribute,
    parse_list_attr_and_index,
)
from pydase.utils.serializer import (
    dump,
    generate_serialized_data_paths,
    get_nested_dict_by_path,
    serialized_dict_is_nested_object,
)

if TYPE_CHECKING:
    from pydase import DataService

logger = logging.getLogger(__name__)


def load_state(func: Callable[..., Any]) -> Callable[..., Any]:
    """This function should be used as a decorator on property setters to indicate that
    the value should be loaded from the JSON file.

    Example:
    >>>     class Service(pydase.DataService):
    ...         _name = "Service"
    ...
    ...         @property
    ...         def name(self) -> str:
    ...             return self._name
    ...
    ...         @name.setter
    ...         @load_state
    ...         def name(self, value: str) -> None:
    ...             self._name = value
    """

    func._load_state = True  # type: ignore[attr-defined]
    return func


def has_load_state_decorator(prop: property) -> bool:
    """Determines if the property's setter method is decorated with the `@load_state`
    decorator.
    """

    try:
        return prop.fset._load_state  # type: ignore[union-attr]
    except AttributeError:
        return False


class StateManager:
    """
    Manages the state of a DataService instance, serving as both a cache and a
    persistence layer. It is designed to provide quick access to the latest known state
    for newly connecting web clients without the need for expensive property accesses
    that may involve complex calculations or I/O operations.

    The StateManager listens for state change notifications from the DataService's
    callback manager and updates its cache accordingly. This cache does not always
    reflect the most current complex property states but rather retains the value from
    the last known state, optimizing for performance and reducing the load on the
    system.

    While the StateManager ensures that the cached state is as up-to-date as possible,
    it does not autonomously update complex properties of the DataService. Such
    properties must be updated programmatically, for instance, by invoking specific
    tasks or methods that trigger the necessary operations to refresh their state.

    The cached state maintained by the StateManager is particularly useful for web
    clients that connect to the system and need immediate access to the current state of
    the DataService. By avoiding direct and potentially costly property accesses, the
    StateManager provides a snapshot of the DataService's state that is sufficiently
    accurate for initial rendering and interaction.

    Attributes:
        cache (dict[str, Any]):
            A dictionary cache of the DataService's state.
        filename (str):
            The file name used for storing the DataService's state.
        service (DataService):
            The DataService instance whose state is being managed.

    Note:
        The StateManager's cache updates are triggered by notifications and do not
        include autonomous updates of complex DataService properties, which must be
        managed programmatically. The cache serves the purpose of providing immediate
        state information to web clients, reflecting the state after the last property
        update.
    """

    def __init__(
        self, service: "DataService", filename: str | Path | None = None
    ) -> None:
        self.filename = getattr(service, "_filename", None)

        if filename is not None:
            if self.filename is not None:
                logger.warning(
                    "Overwriting filename '%s' with '%s'.", self.filename, filename
                )
            self.filename = filename

        self.service = service
        self._data_service_cache = DataServiceCache(self.service)

    @property
    def cache(self) -> dict[str, Any]:
        """Returns the cached DataService state."""
        return self._data_service_cache.cache

    def save_state(self) -> None:
        """
        Saves the DataService's current state to a JSON file defined by `self.filename`.
        Logs an error if `self.filename` is not set.
        """

        if self.filename is not None:
            with open(self.filename, "w") as f:
                json.dump(self.cache["value"], f, indent=4)
        else:
            logger.info(
                "State manager was not initialised with a filename. Skipping "
                "'save_state'..."
            )

    def load_state(self) -> None:
        """
        Loads the DataService's state from a JSON file defined by `self.filename`.
        Updates the service's attributes, respecting type and read-only constraints.
        """

        # Traverse the serialized representation and set the attributes of the class
        json_dict = self._get_state_dict_from_json_file()
        if json_dict == {}:
            logger.debug("Could not load the service state.")
            return

        for path in generate_serialized_data_paths(json_dict):
            nested_json_dict = get_nested_dict_by_path(json_dict, path)
            nested_class_dict = self._data_service_cache.get_value_dict_from_cache(path)

            value, value_type = nested_json_dict["value"], nested_json_dict["type"]
            class_attr_value_type = nested_class_dict.get("type", None)

            if class_attr_value_type == value_type:
                if self.__is_loadable_state_attribute(path):
                    self.set_service_attribute_value_by_path(path, value)
            else:
                logger.info(
                    "Attribute type of '%s' changed from '%s' to "
                    "'%s'. Ignoring value from JSON file...",
                    path,
                    value_type,
                    class_attr_value_type,
                )

    def _get_state_dict_from_json_file(self) -> dict[str, Any]:
        if self.filename is not None and os.path.exists(self.filename):
            with open(self.filename) as f:
                # Load JSON data from file and update class attributes with these
                # values
                return cast(dict[str, Any], json.load(f))
        return {}

    def set_service_attribute_value_by_path(
        self,
        path: str,
        value: Any,
    ) -> None:
        """
        Sets the value of an attribute in the service managed by the `StateManager`
        given its path as a dot-separated string.

        This method updates the attribute specified by 'path' with 'value' only if the
        attribute is not read-only and the new value differs from the current one.
        It also handles type-specific conversions for the new value before setting it.

        Args:
            path: A dot-separated string indicating the hierarchical path to the
                attribute.
            value: The new value to set for the attribute.
        """

        current_value_dict = get_nested_dict_by_path(self.cache["value"], path)

        # This will also filter out methods as they are 'read-only'
        if current_value_dict["readonly"]:
            logger.debug("Attribute '%s' is read-only. Ignoring new value...", path)
            return

        converted_value = self.__convert_value_if_needed(value, current_value_dict)

        # only set value when it has changed
        if self.__attr_value_has_changed(converted_value, current_value_dict["value"]):
            self.__update_attribute_by_path(path, converted_value)
        else:
            logger.debug("Value of attribute '%s' has not changed...", path)

    def __attr_value_has_changed(self, value_object: Any, current_value: Any) -> bool:
        """Check if the serialized value of `value_object` differs from `current_value`.

        The method serializes `value_object` to compare it, which is mainly
        necessary for handling Quantity objects.
        """

        return dump(value_object)["value"] != current_value

    def __convert_value_if_needed(
        self, value: Any, current_value_dict: dict[str, Any]
    ) -> Any:
        if current_value_dict["type"] == "Quantity":
            return u.convert_to_quantity(value, current_value_dict["value"]["unit"])
        if current_value_dict["type"] == "float" and not isinstance(value, float):
            return float(value)
        return value

    def __update_attribute_by_path(self, path: str, value: Any) -> None:
        parent_path_list, attr_name = path.split(".")[:-1], path.split(".")[-1]

        # If attr_name corresponds to a list entry, extract the attr_name and the
        # index
        attr_name, index = parse_list_attr_and_index(attr_name)

        # Update path to reflect the attribute without list indices
        path = ".".join([*parent_path_list, attr_name])

        attr_cache_type = get_nested_dict_by_path(self.cache["value"], path)["type"]

        # Traverse the object according to the path parts
        target_obj = get_object_attr_from_path_list(self.service, parent_path_list)

        if attr_cache_type in ("ColouredEnum", "Enum"):
            enum_attr = get_object_attr_from_path_list(target_obj, [attr_name])
            setattr(target_obj, attr_name, enum_attr.__class__[value])
        elif attr_cache_type == "list":
            list_obj = get_object_attr_from_path_list(target_obj, [attr_name])
            list_obj[index] = value
        else:
            setattr(target_obj, attr_name, value)

    def __is_loadable_state_attribute(self, full_access_path: str) -> bool:
        """Checks if an attribute defined by a dot-separated path should be loaded from
        storage.

        For properties, it verifies the presence of the '@load_state' decorator. Regular
        attributes default to being loadable.
        """

        parent_object = get_object_attr_from_path_list(
            self.service, full_access_path.split(".")[:-1]
        )
        attr_name = full_access_path.split(".")[-1]

        if is_property_attribute(parent_object, attr_name):
            prop = getattr(type(parent_object), attr_name)
            has_decorator = has_load_state_decorator(prop)
            if not has_decorator:
                logger.debug(
                    "Property '%s' has no '@load_state' decorator. "
                    "Ignoring value from JSON file...",
                    attr_name,
                )
            return has_decorator

        cached_serialization_dict = get_nested_dict_by_path(
            self.cache["value"], full_access_path
        )

        if cached_serialization_dict["value"] == "method":
            return False

        # nested objects cannot be loaded
        return not serialized_dict_is_nested_object(cached_serialization_dict)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydase.observer_pattern.observable.observable import Observable

if TYPE_CHECKING:
    from pydase.data_service.data_service import DataService
    from pydase.data_service.task_manager import TaskManager


class AbstractDataService(Observable):
    __root__: DataService
    _task_manager: TaskManager
    _autostart_tasks: dict[str, tuple[Any]]



import logging
from abc import ABC, abstractmethod
from typing import Any

from pydase.observer_pattern.observable import Observable

logger = logging.getLogger(__name__)


class Observer(ABC):
    def __init__(self, observable: Observable) -> None:
        self.observable = observable
        self.observable.add_observer(self)
        self.changing_attributes: list[str] = []

    def _notify_changed(self, changed_attribute: str, value: Any) -> None:
        self.on_change(full_access_path=changed_attribute, value=value)

        if changed_attribute in self.changing_attributes:
            self.changing_attributes.remove(changed_attribute)

    def _notify_change_start(self, changing_attribute: str) -> None:
        self.changing_attributes.append(changing_attribute)
        self.on_change_start(changing_attribute)

    @abstractmethod
    def on_change(self, full_access_path: str, value: Any) -> None:
        ...

    def on_change_start(self, full_access_path: str) -> None:
        return



import asyncio
import json
import logging
from pathlib import Path
from typing import Any

import socketio  # type: ignore[import-untyped]
import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles

from pydase.config import ServiceConfig, WebServerConfig
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.server.web_server.sio_setup import (
    setup_sio_server,
)
from pydase.utils.serializer import generate_serialized_data_paths
from pydase.version import __version__

logger = logging.getLogger(__name__)


class WebServer:
    """
    Represents a web server that adheres to the AdditionalServerProtocol, designed to
    work with a DataService instance. This server facilitates client-server
    communication and state management through web protocols and socket connections.

    The WebServer class initializes and manages a web server environment using FastAPI
    and Socket.IO, allowing for HTTP and WebSocket communications. It incorporates CORS
    (Cross-Origin Resource Sharing) support, custom CSS, and serves a frontend static
    files directory. It also initializes web server settings based on configuration
    files or generates default settings if necessary.

    Configuration for the web server (like service configuration directory and whether
    to generate new web settings) is determined in the following order of precedence:
    1. Values provided directly to the constructor.
    2. Environment variable settings (via configuration classes like
      `pydase.config.ServiceConfig` and `pydase.config.WebServerConfig`).
    3. Default values defined in the configuration classes.

    Args:
        data_service_observer (DataServiceObserver): Observer for the DataService,
          handling state updates and communication to connected clients.
        host (str): Hostname or IP address where the server is accessible. Commonly
          '0.0.0.0' to bind to all network interfaces.
        port (int): Port number on which the server listens. Typically in the range
          1024-65535 (non-standard ports).
        css (str | Path | None, optional): Path to a custom CSS file for styling the
          frontend. If None, no custom styles are applied. Defaults to None.
        enable_cors (bool, optional): Flag to enable or disable CORS policy. When True,
          CORS is enabled, allowing cross-origin requests. Defaults to True.
        config_dir (Path | None, optional): Path to the configuration
          directory where the web settings will be stored. Defaults to
          `pydase.config.ServiceConfig().config_dir`.
        generate_new_web_settings (bool | None, optional): Flag to enable or disable
          generation of new web settings if the configuration file is missing. Defaults
          to `pydase.config.WebServerConfig().generate_new_web_settings`.
        **kwargs (Any): Additional unused keyword arguments.
    """

    def __init__(  # noqa: PLR0913
        self,
        data_service_observer: DataServiceObserver,
        host: str,
        port: int,
        css: str | Path | None = None,
        enable_cors: bool = True,
        config_dir: Path = ServiceConfig().config_dir,
        generate_web_settings: bool = WebServerConfig().generate_web_settings,
        **kwargs: Any,
    ) -> None:
        self.observer = data_service_observer
        self.state_manager = self.observer.state_manager
        self.service = self.state_manager.service
        self.port = port
        self.host = host
        self.css = css
        self.enable_cors = enable_cors
        self._service_config_dir = config_dir
        self._generate_web_settings = generate_web_settings
        self._loop: asyncio.AbstractEventLoop
        self._initialise_configuration()

    async def serve(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._setup_socketio()
        self._setup_fastapi_app()
        self.web_server = uvicorn.Server(
            uvicorn.Config(self.__fastapi_app, host=self.host, port=self.port)
        )
        # overwrite uvicorn's signal handlers, otherwise it will bogart SIGINT and
        # SIGTERM, which makes it impossible to escape out of
        self.web_server.install_signal_handlers = lambda: None  # type: ignore[method-assign]
        await self.web_server.serve()

    def _initialise_configuration(self) -> None:
        logger.debug("Initialising web server configuration...")

        file_path = self._service_config_dir / "web_settings.json"

        if self._generate_web_settings:
            # File does not exist, create it with default content
            logger.debug("Generating web settings file...")
            file_path.parent.mkdir(
                parents=True, exist_ok=True
            )  # Ensure directory exists
            file_path.write_text(json.dumps(self.web_settings, indent=4))

    def _get_web_settings_from_file(self) -> dict[str, dict[str, Any]]:
        file_path = self._service_config_dir / "web_settings.json"
        web_settings = {}

        # File exists, read its content
        if file_path.exists():
            logger.debug(
                "Reading configuration from file '%s' ...", file_path.absolute()
            )

            web_settings = json.loads(file_path.read_text())

        return web_settings

    @property
    def web_settings(self) -> dict[str, dict[str, Any]]:
        current_web_settings = self._get_web_settings_from_file()
        for path in generate_serialized_data_paths(self.state_manager.cache["value"]):
            if path in current_web_settings:
                continue

            current_web_settings[path] = {"displayName": path.split(".")[-1]}

        return current_web_settings

    def _setup_socketio(self) -> None:
        self._sio = setup_sio_server(self.observer, self.enable_cors, self._loop)
        self.__sio_app = socketio.ASGIApp(self._sio)

    def _setup_fastapi_app(self) -> None:  # noqa: C901
        app = FastAPI()

        if self.enable_cors:
            app.add_middleware(
                CORSMiddleware,
                allow_credentials=True,
                allow_origins=["*"],
                allow_methods=["*"],
                allow_headers=["*"],
            )
        app.mount("/ws", self.__sio_app)

        @app.get("/version")
        def version() -> str:
            return __version__

        @app.get("/name")
        def name() -> str:
            return type(self.service).__name__

        @app.get("/service-properties")
        def service_properties() -> dict[str, Any]:
            return self.state_manager.cache

        @app.get("/web-settings")
        def web_settings() -> dict[str, Any]:
            return self.web_settings

        # exposing custom.css file provided by user
        if self.css is not None:

            @app.get("/custom.css")
            async def styles() -> FileResponse:
                return FileResponse(str(self.css))

        app.mount(
            "/",
            StaticFiles(
                directory=Path(__file__).parent.parent.parent / "frontend",
                html=True,
            ),
        )

        self.__fastapi_app = app



import asyncio
import logging
import sys
from copy import copy

import socketio  # type: ignore[import-untyped]
import uvicorn.logging
from uvicorn.config import LOGGING_CONFIG

import pydase.config


class DefaultFormatter(uvicorn.logging.ColourizedFormatter):
    """
    A custom log formatter class that:

    * Outputs the LOG_LEVEL with an appropriate color.
    * If a log call includes an `extras={"color_message": ...}` it will be used
      for formatting the output, instead of the plain text message.
    """

    def formatMessage(self, record: logging.LogRecord) -> str:  # noqa: N802
        recordcopy = copy(record)
        levelname = recordcopy.levelname
        seperator = " " * (8 - len(recordcopy.levelname))
        if self.use_colors:
            levelname = self.color_level_name(levelname, recordcopy.levelno)
            if "color_message" in recordcopy.__dict__:
                recordcopy.msg = recordcopy.__dict__["color_message"]
                recordcopy.__dict__["message"] = recordcopy.getMessage()
        recordcopy.__dict__["levelprefix"] = levelname + seperator
        return logging.Formatter.formatMessage(self, recordcopy)

    def should_use_colors(self) -> bool:
        return sys.stderr.isatty()


class SocketIOHandler(logging.Handler):
    """
    Custom logging handler that emits ERROR and CRITICAL log records to a Socket.IO
    server, allowing for real-time logging in applications that use Socket.IO for
    communication.
    """

    def __init__(self, sio: socketio.AsyncServer) -> None:
        super().__init__(logging.ERROR)
        self._sio = sio

    def format(self, record: logging.LogRecord) -> str:
        return f"{record.name}:{record.funcName}:{record.lineno} - {record.message}"

    def emit(self, record: logging.LogRecord) -> None:
        log_entry = self.format(record)

        loop = asyncio.get_event_loop()
        loop.create_task(
            self._sio.emit(
                "log",
                {
                    "levelname": record.levelname,
                    "message": log_entry,
                },
            )
        )


def setup_logging(level: str | int | None = None) -> None:
    """
    Configures the logging settings for the application.

    This function sets up logging with specific formatting and colorization of log
    messages. The log level is determined based on the application's operation mode,
    with an option to override the level. By default, in a development environment, the
    log level is set to DEBUG, whereas in other environments, it is set to INFO.

    Args:
        level (Optional[str | int]):
            A specific log level to set for the application. If None, the log level is
            determined based on the application's operation mode. Accepts standard log
            level names ('DEBUG', 'INFO', etc.) and corresponding numerical values.

    Example:

    ```python
    >>> import logging
    >>> setup_logging(logging.DEBUG)
    >>> setup_logging("INFO")
    ```
    """

    logger = logging.getLogger()

    if pydase.config.OperationMode().environment == "development":
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    # If a level is specified, check whether it's a string or an integer.
    if level is not None:
        if isinstance(level, str):
            # Convert known log level strings directly to their corresponding logging
            # module constants.
            level_name = level.upper()  # Ensure level names are uppercase
            if hasattr(logging, level_name):
                log_level = getattr(logging, level_name)
            else:
                raise ValueError(
                    f"Invalid log level: {level}. Must be one of 'DEBUG', 'INFO', "
                    "'WARNING', 'ERROR', etc."
                )
        elif isinstance(level, int):
            log_level = level  # Directly use integer levels
        else:
            raise ValueError("Log level must be a string or an integer.")

    # Set the logger's level.
    logger.setLevel(log_level)

    # create console handler and set level to debug
    ch = logging.StreamHandler()

    # add formatter to ch
    ch.setFormatter(
        DefaultFormatter(
            fmt=(
                "%(asctime)s.%(msecs)03d | %(levelprefix)s | "
                "%(name)s:%(funcName)s:%(lineno)d - %(message)s"
            ),
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )

    # add ch to logger
    logger.addHandler(ch)

    logger.debug("Configuring service logging.")
    logging.getLogger("asyncio").setLevel(logging.INFO)
    logging.getLogger("urllib3").setLevel(logging.INFO)

    # configuring uvicorn logger
    LOGGING_CONFIG["formatters"]["default"][
        "fmt"
    ] = "%(asctime)s.%(msecs)03d | %(levelprefix)s %(message)s"
    LOGGING_CONFIG["formatters"]["default"]["datefmt"] = "%Y-%m-%d %H:%M:%S"
    LOGGING_CONFIG["formatters"]["access"]["fmt"] = (
        "%(asctime)s.%(msecs)03d | %(levelprefix)s %(client_addr)s "
        '- "%(request_line)s" %(status_code)s'
    )
    LOGGING_CONFIG["formatters"]["access"]["datefmt"] = "%Y-%m-%d %H:%M:%S"



import asyncio

import pydase


class DeviceConnection(pydase.DataService):
    """
    Base class for device connection management within the pydase framework.

    This class serves as the foundation for subclasses that manage connections to
    specific devices. It implements automatic reconnection logic that periodically
    checks the device's availability and attempts to reconnect if the connection is
    lost. The frequency of these checks is controlled by the `_reconnection_wait_time`
    attribute.

    Subclassing
    -----------
    Users should primarily override the `connect` method to establish a connection
    to the device. This method should update the `self._connected` attribute to reflect
    the connection status:

    >>> class MyDeviceConnection(DeviceConnection):
    ...     def connect(self) -> None:
    ...         # Implementation to connect to the device
    ...         # Update self._connected to `True` if connection is successful,
    ...         # `False` otherwise
    ...         ...

    Optionally, if additional logic is needed to determine the connection status,
    the `connected` property can also be overridden:

    >>> class MyDeviceConnection(DeviceConnection):
    ...     @property
    ...     def connected(self) -> bool:
    ...         # Custom logic to determine connection status
    ...         return some_custom_condition
    ...

    Frontend Representation
    -----------------------
    In the frontend, this class is represented without directly exposing the `connect`
    method and `connected` attribute. Instead, user-defined attributes, methods, and
    properties are displayed. When `self.connected` is `False`, the frontend component
    shows an overlay that allows manual triggering of the `connect()` method. This
    overlay disappears once the connection is successfully re-established.
    """

    def __init__(self) -> None:
        super().__init__()
        self._connected = False
        self._autostart_tasks["_handle_connection"] = ()  # type: ignore
        self._reconnection_wait_time = 10.0

    def connect(self) -> None:
        """Tries connecting to the device and changes `self._connected` status
        accordingly. This method is called every `self._reconnection_wait_time` seconds
        when `self.connected` is False. Users should override this method to implement
        device-specific connection logic.
        """

    @property
    def connected(self) -> bool:
        """Indicates if the device is currently connected or was recently connected.
        Users may override this property to incorporate custom logic for determining
        the connection status.
        """
        return self._connected

    async def _handle_connection(self) -> None:
        """Automatically tries reconnecting to the device if it is not connected.
        This method leverages the `connect` method and the `connected` property to
        manage the connection status.
        """
        while True:
            if not self.connected:
                self.connect()
            await asyncio.sleep(self._reconnection_wait_time)



import logging
from typing import Any

from pydase.data_service.data_service import DataService

logger = logging.getLogger(__name__)


class NumberSlider(DataService):
    """
    This class models a UI slider for a data service, allowing for adjustments of a
    parameter within a specified range and increments.

    Parameters:
    -----------
    value (float, optional):
        The initial value of the slider. Defaults to 0.
    min (float, optional):
        The minimum value of the slider. Defaults to 0.
    max (float, optional):
        The maximum value of the slider. Defaults to 100.
    step_size (float, optional):
        The increment/decrement step size of the slider. Defaults to 1.0.

    Example:
    --------
    ```python
    class MySlider(pydase.components.NumberSlider):
        def __init__(
            self,
            value: float = 0.0,
            min_: float = 0.0,
            max_: float = 100.0,
            step_size: float = 1.0,
        ) -> None:
            super().__init__(value, min_, max_, step_size)

        @property
        def min(self) -> float:
            return self._min

        @min.setter
        def min(self, value: float) -> None:
            self._min = value

        @property
        def max(self) -> float:
            return self._max

        @max.setter
        def max(self, value: float) -> None:
            self._max = value

        @property
        def step_size(self) -> float:
            return self._step_size

        @step_size.setter
        def step_size(self, value: float) -> None:
            self._step_size = value

        @property
        def value(self) -> float:
            return self._value

        @value.setter
        def value(self, value: float) -> None:
            if value < self._min or value > self._max:
                raise ValueError(
                    "Value is either below allowed min or above max value."
                )

            self._value = value

    class MyService(pydase.DataService):
        def __init__(self) -> None:
            self.voltage = MyService()

    # Modifying or accessing the voltage value:
    my_service = MyService()
    my_service.voltage.value = 5
    print(my_service.voltage.value)  # Output: 5
    ```
    """

    def __init__(
        self,
        value: Any = 0.0,
        min_: float = 0.0,
        max_: float = 100.0,
        step_size: float = 1.0,
    ) -> None:
        super().__init__()
        self._step_size = step_size
        self._value = value
        self._min = min_
        self._max = max_

    @property
    def min(self) -> float:
        """The min property."""
        return self._min

    @property
    def max(self) -> float:
        """The min property."""
        return self._max

    @property
    def step_size(self) -> float:
        """The min property."""
        return self._step_size

    @property
    def value(self) -> Any:
        """The value property."""
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        self._value = value
